@@ -1,5 +1,6 @@
 from abc import ABC, abstractmethod
-from numpy import abs, exp, eye, log, zeros
+from inspect import isclass
+from numpy import abs, exp, eye, log, zeros, ndarray
 
 
 class CovarianceFunction(ABC):
@@ -8,7 +9,11 @@ class CovarianceFunction(ABC):
     """
 
     @abstractmethod
-    def pass_data(self, x, y):
+    def pass_spatial_data(self, x):
+        pass
+
+    @abstractmethod
+    def estimate_hyperpar_bounds(self, y):
         pass
 
     @abstractmethod
@@ -40,28 +45,38 @@ class CovarianceFunction(ABC):
 class CompositeCovariance(CovarianceFunction):
     def __init__(self, covariance_components):
         self.components = covariance_components
+        self.bounds = None
 
-    def pass_data(self, x, y):
-        [comp.pass_data(x, y) for comp in self.components]
-        # here we'd need to build slices to pass parameters
-        # to each component, and construct a top-level bounds list
-        self.slices = []
-        for comp in self.components:
-            L = comp.n_params
-            if len(self.slices) == 0:
-                self.slices.append(slice(0, L))
-            else:
-                last = self.slices[-1].stop
-                self.slices.append(slice(last, last + L))
-
-        self.bounds = []
-        [self.bounds.extend(comp.bounds) for comp in self.components]
-
+    def pass_spatial_data(self, x):
+        """
+        Pre-calculates hyperparameter-independent part of the data covariance
+        matrix as an optimisation.
+        """
+        [comp.pass_spatial_data(x) for comp in self.components]
+        # Create slices to address the parameters of each component
+        self.slices = slice_builder([c.n_params for c in self.components])
+        # combine hyperparameter labels for each component
         self.hyperpar_labels = []
         for i, comp in enumerate(self.components):
-            labels = [f"K{i+1}_{s}" for s in comp.hyperpar_labels]
+            labels = [f"K{i+1}: {s}" for s in comp.hyperpar_labels]
             self.hyperpar_labels.extend(labels)
-        self.n_params = len(self.hyperpar_labels)
+        # check for consistency of length of bounds, labels
+        self.n_params = sum(c.n_params for c in self.components)
+        assert self.n_params == len(self.hyperpar_labels)
+
+    def estimate_hyperpar_bounds(self, y):
+        """
+        Estimates bounds on the hyper-parameters to be
+        used during optimisation.
+        """
+        # estimate bounds for components where they were not specified
+        for comp in self.components:
+            if comp.bounds is None:
+                comp.estimate_hyperpar_bounds(y)
+        # combine parameter bounds for each component
+        self.bounds = []
+        [self.bounds.extend(comp.bounds) for comp in self.components]
+        assert self.n_params == len(self.bounds)
 
     def __call__(self, u, v, theta):
         return sum(
@@ -108,7 +123,7 @@ class WhiteNoise(CovarianceFunction):
        from inference.gp import SquaredExponential, WhiteNoise
        composite_kernel = SquaredExponential() + WhiteNoise()
 
-    :param hyperpar_bounds: \
+    :param hyperpar_bounds:
         By default, ``WhiteNoise`` will automatically set sensible lower and
         upper bounds on the value of the log-noise-level based on the available data.
         However, this keyword allows the bounds to be specified manually as a length-2
@@ -117,21 +132,24 @@ class WhiteNoise(CovarianceFunction):
 
     def __init__(self, hyperpar_bounds=None):
         self.bounds = hyperpar_bounds
+        self.n_params = 1
+        self.hyperpar_labels = ["WhiteNoise log-sigma"]
 
-    def pass_data(self, x, y):
+    def pass_spatial_data(self, x: ndarray):
         """
         Pre-calculates hyperparameter-independent part of the data covariance
-        matrix as an optimisation, and sets bounds on hyperparameter values.
+        matrix as an optimisation.
         """
-        self.I = eye(y.size)
+        self.I = eye(x.shape[0])
 
+    def estimate_hyperpar_bounds(self, y: ndarray):
+        """
+        Estimates bounds on the hyper-parameters to be
+        used during optimisation.
+        """
         # construct sensible bounds on the hyperparameter values
-        if self.bounds is None:
-            s = log(y.ptp())
-            self.bounds = [(s - 8, s + 2)]
-
-        self.n_params = 1
-        self.hyperpar_labels = ["log_noise_level"]
+        s = log(y.ptp())
+        self.bounds = [(s - 8, s + 2)]
 
     def __call__(self, u, v, theta):
         return zeros([u.size, v.size])
@@ -171,7 +189,7 @@ class SquaredExponential(CovarianceFunction):
 
        \underline{\theta} = [ \ln{A}, \ln{l_1}, \ldots, \ln{l_n}]
 
-    :param hyperpar_bounds: \
+    :param hyperpar_bounds:
         By default, ``SquaredExponential`` will automatically set sensible lower and
         upper bounds on the value of the hyperparameters based on the available data.
         However, this keyword allows the bounds to be specified manually as a list of
@@ -180,29 +198,38 @@ class SquaredExponential(CovarianceFunction):
 
     def __init__(self, hyperpar_bounds=None):
         self.bounds = hyperpar_bounds
+        self.n_params: int
+        self.dx: ndarray
+        self.distances: ndarray
+        self.hyperpar_labels: list
 
-    def pass_data(self, x, y):
+    def pass_spatial_data(self, x: ndarray):
         """
         Pre-calculates hyperparameter-independent part of the data covariance
-        matrix as an optimisation, and sets bounds on hyperparameter values.
+        matrix as an optimisation.
         """
         # distributed outer subtraction using broadcasting
-        dx = x[:, None, :] - x[None, :, :]
-        self.distances = -0.5 * dx**2
+        self.dx = x[:, None, :] - x[None, :, :]
+        self.distances = -0.5 * self.dx**2
         # small values added to the diagonal for stability
-        self.epsilon = 1e-12 * eye(dx.shape[0])
-
-        # construct sensible bounds on the hyperparameter values
-        if self.bounds is None:
-            s = log(y.std())
-            self.bounds = [(s - 4, s + 4)]
-            for i in range(x.shape[1]):
-                lwr = log(abs(dx[:, :, i]).mean()) - 4
-                upr = log(dx[:, :, i].max()) + 2
-                self.bounds.append((lwr, upr))
+        self.epsilon = 1e-12 * eye(self.dx.shape[0])
         self.n_params = x.shape[1] + 1
-        self.hyperpar_labels = ["log_amplitude"]
-        self.hyperpar_labels.extend([f"log_scale_{i}" for i in range(x.shape[1])])
+        self.hyperpar_labels = ["SqrExp log-amplitude"]
+        self.hyperpar_labels.extend(
+            [f"SqrExp log-scale {i}" for i in range(x.shape[1])]
+        )
+
+    def estimate_hyperpar_bounds(self, y: ndarray):
+        """
+        Estimates bounds on the hyper-parameters to be
+        used during optimisation.
+        """
+        s = log(y.std())
+        self.bounds = [(s - 4, s + 4)]
+        for i in range(self.dx.shape[2]):
+            lwr = log(abs(self.dx[:, :, i]).mean()) - 4
+            upr = log(self.dx[:, :, i].max()) + 2
+            self.bounds.append((lwr, upr))
 
     def __call__(self, u, v, theta):
         a = exp(theta[0])
@@ -263,7 +290,7 @@ class RationalQuadratic(CovarianceFunction):
 
        \underline{\theta} = [ \ln{A}, \ln{\alpha}, \ln{l_1}, \ldots, \ln{l_n}]
 
-    :param hyperpar_bounds: \
+    :param hyperpar_bounds:
         By default, ``RationalQuadratic`` will automatically set sensible lower and
         upper bounds on the value of the hyperparameters based on the available data.
         However, this keyword allows the bounds to be specified manually as a list of
@@ -273,28 +300,31 @@ class RationalQuadratic(CovarianceFunction):
     def __init__(self, hyperpar_bounds=None):
         self.bounds = hyperpar_bounds
 
-    def pass_data(self, x, y):
+    def pass_spatial_data(self, x: ndarray):
         """
         Pre-calculates hyperparameter-independent part of the data covariance
-        matrix as an optimisation, and sets bounds on hyperparameter values.
+        matrix as an optimisation.
         """
         # distributed outer subtraction using broadcasting
-        dx = x[:, None, :] - x[None, :, :]
-        self.distances = 0.5 * dx**2
+        self.dx = x[:, None, :] - x[None, :, :]
+        self.distances = 0.5 * self.dx**2
         # small values added to the diagonal for stability
-        self.epsilon = 1e-12 * eye(dx.shape[0])
-
-        # construct sensible bounds on the hyperparameter values
-        if self.bounds is None:
-            s = log(y.std())
-            self.bounds = [(s - 4, s + 4), (-2, 6)]
-            for i in range(x.shape[1]):
-                lwr = log(abs(dx[:, :, i]).mean()) - 4
-                upr = log(dx[:, :, i].max()) + 2
-                self.bounds.append((lwr, upr))
+        self.epsilon = 1e-12 * eye(self.dx.shape[0])
         self.n_params = x.shape[1] + 2
-        self.hyperpar_labels = ["log_amplitude", "log_alpha"]
-        self.hyperpar_labels.extend([f"log_scale_{i}" for i in range(x.shape[1])])
+        self.hyperpar_labels = ["RQ log-amplitude", "RQ log-alpha"]
+        self.hyperpar_labels.extend([f"RQ log-scale {i}" for i in range(x.shape[1])])
+
+    def estimate_hyperpar_bounds(self, y: ndarray):
+        """
+        Estimates bounds on the hyper-parameters to be
+        used during optimisation.
+        """
+        s = log(y.std())
+        self.bounds = [(s - 4, s + 4), (-2, 6)]
+        for i in range(self.dx.shape[2]):
+            lwr = log(abs(self.dx[:, :, i]).mean()) - 4
+            upr = log(self.dx[:, :, i].max()) + 2
+            self.bounds.append((lwr, upr))
 
     def __call__(self, u, v, theta):
         a = exp(theta[0])
@@ -330,3 +360,165 @@ class RationalQuadratic(CovarianceFunction):
 
     def get_bounds(self):
         return self.bounds
+
+
+class ChangePoint(CovarianceFunction):
+    r"""
+    ``ChangePoint`` is a covariance function which divides the input space into two
+    regions (at some point along a chosen input dimension), allowing each of the two
+    regions to be modelled using a separate covariance function.
+
+    This is useful in cases where properties of the data (e.g. the scale-lengths
+    over which the data vary) change significantly over the input dimension which is
+    used to divide the space.
+
+    The change-point kernel :math:`K_{\mathrm{cp}}` is a weighted-sum of the two
+    input kernels :math:`K_{1}, \, K_{2}` which model each of the two regions:
+
+    .. math::
+
+       K_{\mathrm{cp}}(u, v) = K_{1}(u, v) (1 - f(u))(1 - f(v)) + K_{2}(u, v) f(u) f(v)
+
+    where the weighting :math:`f(x)` is the logistic function
+
+    .. math::
+
+       f(x) = \frac{1}{1 + e^{-(x - x_0) / w}}
+
+    and :math:`x_0, \, w` are the location and width of the change-point respectively.
+    :math:`x_0` and :math:`w` are hyperparameters which are determined automatically
+    (alongside the hyperparameters for :math:`K_{1}, \, K_{2}`).
+
+    :param K1:
+        The covariance kernel which applies to the 'low' side of the change-point.
+
+    :param K2:
+        The covariance kernel which applies to the 'high' side of the change-point.
+
+    :param int axis:
+        The spatial axis over which the transition between the two kernels occurs.
+
+    :param location_bounds:
+        The bounds for the change-point location hyperparameter
+        :math:`x_0` as a tuple of the form ``(lower_bound, upper_bound)``.
+
+    :param width_bounds:
+        The bounds for the change-point width hyperparameter :math:`w` as a tuple
+        of the form ``(lower_bound, upper_bound)``.
+    """
+
+    def __init__(
+        self,
+        K1=SquaredExponential,
+        K2=SquaredExponential,
+        axis=0,
+        location_bounds=None,
+        width_bounds=None,
+    ):
+        self.cov1 = K1() if isclass(K1) else K1
+        self.cov2 = K2() if isclass(K2) else K2
+        self.axis = axis
+        self.location_bounds = check_bounds(location_bounds)
+        self.width_bounds = check_bounds(width_bounds)
+        self.hyperpar_labels = []
+        self.bounds = []
+
+    def pass_spatial_data(self, x: ndarray):
+        self.cov1.pass_spatial_data(x)
+        self.cov2.pass_spatial_data(x)
+        # Create slices to address the parameters of each component
+        param_counts = [self.cov1.n_params, self.cov2.n_params, 2]
+        self.n_params = sum(param_counts)
+        self.K1_slc, self.K2_slc, self.CP_slc = slice_builder(param_counts)
+        # combine hyperparameter labels for K1, K2 and the change-point
+        label_groups = [
+            [f"ChngPnt K1: {lab}" for lab in self.cov1.hyperpar_labels],
+            [f"ChngPnt K2: {lab}" for lab in self.cov2.hyperpar_labels],
+            ["ChngPnt location", "ChngPnt width"],
+        ]
+        [self.hyperpar_labels.extend(L) for L in label_groups]
+
+        # store x-data from the dimension of the change-point
+        self.x_cp = x[:, self.axis]
+        assert self.n_params == len(self.hyperpar_labels)
+
+    def estimate_hyperpar_bounds(self, y: ndarray):
+        xr = self.x_cp.min(), self.x_cp.max()
+        dx = xr[1] - xr[0]
+        # combine parameter bounds for K1, K2 and the change-point
+        if self.cov1.bounds is None:
+            self.cov1.estimate_hyperpar_bounds(y)
+        if self.cov2.bounds is None:
+            self.cov2.estimate_hyperpar_bounds(y)
+        self.bounds.extend(self.cov1.bounds)
+        self.bounds.extend(self.cov2.bounds)
+        self.bounds.extend(
+            [
+                xr if self.location_bounds is None else self.location_bounds,
+                (5e-3 * dx, 0.5 * dx)
+                if self.width_bounds is None
+                else self.width_bounds,
+            ]
+        )
+        # check for consistency of length of bounds
+        assert self.n_params == len(self.bounds)
+
+    def __call__(self, u, v, theta):
+        K1 = self.cov1(u, v, theta[self.K1_slc])
+        K2 = self.cov2(u, v, theta[self.K2_slc])
+        w_u = self.logistic(u[:, self.axis], theta[self.CP_slc])
+        w_v = self.logistic(v[:, self.axis], theta[self.CP_slc])
+        w1 = (1 - w_u)[:, None] * (1 - w_v)[None, :]
+        w2 = w_u[:, None] * w_v[None, :]
+        return K1 * w1 + K2 * w2
+
+    def build_covariance(self, theta):
+        K1 = self.cov1.build_covariance(theta[self.K1_slc])
+        K2 = self.cov2.build_covariance(theta[self.K2_slc])
+        w = self.logistic(self.x_cp, theta[self.CP_slc])
+        w1 = (1 - w)[:, None] * (1 - w)[None, :]
+        w2 = w[:, None] * w[None, :]
+        return K1 * w1 + K2 * w2
+
+    def covariance_and_gradients(self, theta):
+        K1, K1_grads = self.cov1.covariance_and_gradients(theta[self.K1_slc])
+        K2, K2_grads = self.cov2.covariance_and_gradients(theta[self.K2_slc])
+        w, w_grads = self.logistic_and_gradient(self.x_cp, theta[self.CP_slc])
+        w1 = (1 - w)[:, None] * (1 - w)[None, :]
+        w2 = w[:, None] * w[None, :]
+        K = K1 * w1 + K2 * w2
+        gradients = [c * w1 for c in K1_grads]
+        gradients.extend([c * w2 for c in K2_grads])
+        for g in w_grads:
+            A = -g[:, None] * (1 - w)[None, :]
+            B = g[:, None] * w[None, :]
+            gradients.append(K1 * (A + A.T) + K2 * (B + B.T))
+        return K, gradients
+
+    @staticmethod
+    def logistic(x, theta):
+        z = (x - theta[0]) / theta[1]
+        return 1.0 / (1.0 + exp(-z))
+
+    @staticmethod
+    def logistic_and_gradient(x, theta):
+        z = (x - theta[0]) / theta[1]
+        f = 1.0 / (1.0 + exp(-z))
+        dfdc = -f * (1 - f) / theta[1]
+        return f, [dfdc, dfdc * z]
+
+
+def slice_builder(lengths):
+    slices = [slice(0, lengths[0])]
+    for L in lengths[1:]:
+        last = slices[-1].stop
+        slices.append(slice(last, last + L))
+    return slices
+
+
+def check_bounds(bounds):
+    if bounds is not None:
+        assert type(bounds) in [list, tuple, ndarray]
+        assert len(bounds) == 2
+        assert bounds[1] > bounds[0]
+    return bounds
